"""Python S/MIME handler."""

__author__ = 'G. B. Versiani'
__license__ = 'Apache License (2.0)'
__version__ = '0.0.1'

from .encrypt import encrypt

__all__ = [encrypt, __author__, __license__, __version__]
